# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the function for evaluating the shape of a line.
"""

from ._run import _evaluate_line



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the function which evaluates the shape of a cluster by delegating to the
implementation for the given cluster dimension.
"""

import warnings

from fsc.export import export

from ._point import _evaluate_point
from ._line import _evaluate_line

_WEIGHT_KEY = '_weight'


@export
def evaluate_cluster(
    graph,
    dim,
    coordinate_system,
    feature_size,
    evaluate_line_method='shortest_path'
):
    """
    Evaluate the shape of a cluster with the given positions.

    Arguments
    ---------
    graph : nx.Graph
        Graph describing the cluster.
    dim : int
        Dimension of the cluster.
    coordinate_system : CoordinateSystem
        Coordinate system used to calculate distances.
    feature_size : float
        Distance between two nodal points at which they are considered distinct.

    Returns
    -------
    :obj:`None` or NodalPoint or NodalLine :
        The shape of the given positions. Returns ``None`` if the shape could
        not be determined.
    """
    if dim == 0:
        return _evaluate_point(
            positions=list(graph.nodes), coordinate_system=coordinate_system
        )
    elif dim == 1:
        try:
            return _evaluate_line(
                graph=graph,
                coordinate_system=coordinate_system,
                feature_size=feature_size,
                method=evaluate_line_method,
            )
        except (IndexError, ValueError) as exc:
            warnings.warn('Could not identify line: {}'.format(exc))
    else:
        return None



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the function which runs the search step.
"""

import queue
import asyncio
import threading
from functools import partial
from types import MappingProxyType

from fsc.export import export

from ._controller import Controller
from ._logging import SEARCH_LOGGER


@export
async def run_async(
    gap_fct,
    *,
    limits=((0, 1), ) * 3,
    periodic=True,
    initial_state=None,
    save_file=None,
    save_delay=5.,
    load=False,
    load_quiet=True,
    initial_mesh_size=10,
    force_initial_mesh=False,
    refinement_stencil='auto',
    gap_threshold=1e-6,
    feature_size=2e-3,
    use_fake_potential=False,
    nelder_mead_kwargs=MappingProxyType({}),
    num_minimize_parallel=50,
    recheck_pos_dist=True,
    recheck_count_cutoff=0,
    simplex_check_cutoff=0
):
    """Run the nodal point search.

    Arguments
    ---------
    gap_fct : collections.abc.Callable
        Function or coroutine describing the potential of which nodes should be
        found.
    limits : tuple(tuple(float))
        The limits of the box where nodes are searched, given as tuple for each
        dimension.
    periodic : bool
        Indicates whether periodic boundary conditions are used for the
        coordinate system.
    save_file : str
        Path to the file where the intermediate results are stored.
    save_delay : float
        Minimum delay (in seconds) between saving the results.
    load : bool
        Enable or disable loading the initial state from ``save_file``.
    load_quiet : bool
        When set to ``True``, ignore errors when loading the initial state.
    initial_mesh_size : int or tuple(int)
        Size of the initial mesh of starting points. Can be given either as a
        single integer or as a list of integers corresponding to the different
        dimensions.
    force_initial_mesh : bool
        If ``True``, add the initial mesh also when restarting the calculation
        from an intermediate result.
    refinement_stencil : np.array
        The stencil of simplices used in the refinement. Normalized such that
        the starting point is at the origin, and the dist_cutoff is one. The
        array must have dimensions (N, dim + 1, dim), where dim is the dimension
        of the problem, and N can be chosen to be the number of refinement
        minimization runs.
    gap_threshold : float
        Threshold for the function value for which a given point is considered
        to be a node.
    feature_size : float
        Threshold for the distance between two nodes where they are considered
        distinct.
    use_fake_potential : bool
        If ``True``, the minimization for a given simplex is performed in two
        steps, first adding a fake potential to repel the minimization from
        existing nodes.
    nelder_mead_kwargs : collections.abc.Mapping
        Keyword arguments passed to the Nelder-Mead algorithm.
    num_minimize_parallel : int
        Maximum number of minimization calculations which are launched in
        parallel.
    recheck_pos_dist : bool
        Indicates whether the position of a refinement box is checked again
        before launching the corresponding refinement.
    recheck_count_cutoff : int
        Number of positions which are allowed to be within the cutoff distance
        when re-checking the position.
    simplex_check_cutoff : int
        Number of vertices which are allowed to be within the cutoff distance
        when re-checking the simplex.

    Returns
    -------
    SearchResultContainer:
        The result of the search algorithm.
    """
    SEARCH_LOGGER.debug('Initializing search controller.')
    controller = Controller(
        gap_fct=gap_fct,
        limits=limits,
        periodic=periodic,
        initial_state=initial_state,
        save_file=save_file,
        save_delay=save_delay,
        load=load,
        load_quiet=load_quiet,
        initial_mesh_size=initial_mesh_size,
        force_initial_mesh=force_initial_mesh,
        gap_threshold=gap_threshold,
        feature_size=feature_size,
        use_fake_potential=use_fake_potential,
        nelder_mead_kwargs=nelder_mead_kwargs,
        num_minimize_parallel=num_minimize_parallel,
        refinement_stencil=refinement_stencil,
        recheck_pos_dist=recheck_pos_dist,
        recheck_count_cutoff=recheck_count_cutoff,
        simplex_check_cutoff=simplex_check_cutoff
    )
    SEARCH_LOGGER.debug('Running search controller.')
    await controller.run()
    SEARCH_LOGGER.debug('Search controller finished.')
    return controller.state.result


@export
def run(*args, **kwargs):
    """Wrapper around :func:`.run_async` that runs the node search synchronously.

    Arguments
    ---------
    args : tuple
        Positional arguments passed to :func:`.run_async`.
    kwargs : collections.abc.Mapping
        Keyword arguments passed to :func:`.run_async`.
    """
    try:
        loop = asyncio.get_event_loop()
        close_loop = False
    except RuntimeError:
        SEARCH_LOGGER.debug('Creating a new event loop.')
        loop = asyncio.new_event_loop()
        close_loop = True

    try:
        if loop.is_running():
            SEARCH_LOGGER.debug('Running in a separate thread.')
            res_queue = queue.Queue()
            exc_queue = queue.Queue()
            thread = threading.Thread(
                target=partial(
                    _run_in_thread,
                    *args,
                    res_queue=res_queue,
                    exc_queue=exc_queue,
                    **kwargs
                )
            )
            SEARCH_LOGGER.debug('Starting thread.')
            thread.start()
            SEARCH_LOGGER.debug('Joining thread.')
            thread.join()
            if not exc_queue.empty():
                raise exc_queue.get()
            res = res_queue.get()
        else:
            SEARCH_LOGGER.debug('Running in the current thread.')
            res = loop.run_until_complete(run_async(*args, **kwargs))
    finally:
        if close_loop:
            loop.close()
    return res


def _run_in_thread(*args, res_queue, exc_queue, **kwargs):
    """
    Helper function that runs the search function, to be used as a thread target.
    This assumes that no (running or other) loop exists.
    """
    try:
        loop = asyncio.new_event_loop()
        res = loop.run_until_complete(run_async(*args, **kwargs))
        loop.close()
        res_queue.put(res)
    except Exception as exc:
        exc_queue.put(exc)
        raise



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Implements line evaluation with the dominating set method.
"""

import operator
import itertools
from collections import namedtuple

import numpy as np
import networkx as nx

from ....search._controller import _DIST_CUTOFF_FACTOR
from ..._cluster import _DISTANCE_KEY
from ..._logging import IDENTIFY_LOGGER

_WEIGHT_KEY = '_weight'


def _evaluate_line_dominating_set(*, graph, coordinate_system, feature_size):
    """
    Evaluate the positions of a nodal line using the 'dominating set' method.
    """
    graph_copy = graph.copy()
    graph_reduced_neighbours = nx.Graph()
    graph_reduced_neighbours.add_nodes_from(graph_copy.nodes)

    for edge in graph_copy.edges:
        edge_attrs = graph_copy.edges[edge]
        distance = edge_attrs[_DISTANCE_KEY]
        edge_attrs[_WEIGHT_KEY] = distance**4
        if distance < feature_size / _DIST_CUTOFF_FACTOR:
            graph_reduced_neighbours.add_edge(*edge)

    dominating_set = nx.algorithms.dominating_set(graph_reduced_neighbours)
    subgraph = graph.subgraph(dominating_set).copy()

    _patch_all_subgraph_holes(
        subgraph=subgraph,
        graph=graph_copy,
        coordinate_system=coordinate_system,
        feature_size=feature_size,
    )
    _remove_duplicate_paths(subgraph)

    return subgraph


def _patch_all_subgraph_holes(
    *,
    subgraph,
    graph,
    coordinate_system,
    feature_size,
    candidates=None,
    weight=_DISTANCE_KEY
):
    """
    Check for 'holes' where the subgraph is disconnected while the original graph is not, and patch them by adding the shortest path on the full graph between the points in the subgraph.
    """
    pair_to_patch = namedtuple('pair_to_patch', ['edge', 'dist_original'])
    to_patch = []

    patch_cutoff = 1.001
    if candidates is None:
        candidates = subgraph.nodes
    for pos1, pos2 in itertools.combinations(candidates, r=2):
        # The minimum distance is used here only to improve performance,
        # because it is much quicker to calculate than the shortest path.
        # By using the distance in the coordinate system as a lower limit for
        # 'dist_original', we can abort early in many cases.
        dist_minimal = coordinate_system.distance(
            np.array(pos1), np.array(pos2)
        )
        if dist_minimal < 2 * feature_size:
            dist_reduced = nx.algorithms.shortest_path_length(
                subgraph, pos1, pos2, weight=weight
            )
            if dist_reduced > max(feature_size, patch_cutoff * dist_minimal):
                dist_original = nx.algorithms.shortest_path_length(
                    graph, pos1, pos2, weight=weight
                )
                if dist_reduced > patch_cutoff * dist_original and dist_original < 2 * feature_size:
                    to_patch.append(
                        pair_to_patch(
                            edge=(pos1, pos2), dist_original=dist_original
                        )
                    )

    to_patch = sorted(to_patch, key=operator.attrgetter('dist_original'))
    for edge, dist_original in to_patch:
        # might have changed since the subgraph is being patched
        dist_reduced = nx.algorithms.shortest_path_length(
            subgraph, *edge, weight=_DISTANCE_KEY
        )
        if dist_reduced > 2 * dist_original:
            IDENTIFY_LOGGER.debug(
                'Patching hole {} in sub-graph.'.format(edge)
            )
            start, end = edge
            _patch_subgraph_hole(
                subgraph=subgraph, graph=graph, start=start, end=end
            )


def _patch_subgraph_hole(*, subgraph, graph, start, end):
    """
    Patch a single hole in the subgraph between the given start and end points.
    """
    shortest_path = nx.algorithms.shortest_path(
        graph, start, end, weight=_DISTANCE_KEY
    )
    for node in shortest_path[1:-1]:
        subgraph.add_node(node)
    for edge in zip(shortest_path[:-1], shortest_path[1:]):
        subgraph.add_edge(*edge, **graph.edges[edge])


def _remove_duplicate_paths(subgraph):
    """
    Remove all "duplicate" edges from the graph where there is another path with lower total weight connecting the two points.
    """
    for *edge, _ in sorted(
        subgraph.edges(data=_WEIGHT_KEY), key=lambda e: -e[2]
    ):
        shortest_path = nx.algorithms.shortest_path(
            subgraph, *edge, weight=_WEIGHT_KEY
        )
        if len(shortest_path) > 2:
            subgraph.remove_edge(*edge)



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the function which evaluates the shape of a point.
"""

import numpy as np

from ..result import NodalPoint


def _evaluate_point(positions, coordinate_system):
    return NodalPoint(
        position=coordinate_system.
        average([np.array(pos) for pos in positions])
    )



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Implements the 'ballistic' method for evaluating the line, which starts at a
specific point and then tries going into the most populated direction. The
angle to the previous step and distance to the next node are also taken into
consideration in choosing each step.
"""

import numpy as np
import scipy.linalg as la
import networkx as nx

from ..._logging import IDENTIFY_LOGGER
from ..._cluster import _DISTANCE_KEY

# Multipliers for the weight which determines the next point to pick
# These are the 'magic numbers' which determine how well the
# line evaluation performs.
_MULTIPLIER_HIGH_DEGREE = 1
_MULTIPLIER_LOW_DEGREE = 1
_MULTIPLIER_EXISTING_NODE = 0.2
_MULTIPLIER_DISTANCE = 5


def _evaluate_line_ballistic(*, graph, coordinate_system, feature_size):
    """
    Runs the line evaluation using the 'ballistic' method.
    """
    runner = _BallisticLineImpl(
        graph=graph,
        coordinate_system=coordinate_system,
        feature_size=feature_size
    )
    runner.run()
    return runner.result_graph


class _BallisticLineImpl:
    """
    Implementation class for the 'ballistic' line evaluation.
    """
    def __init__(self, *, graph, coordinate_system, feature_size):
        self.graph = graph
        self.coordinate_system = coordinate_system
        self.feature_size = feature_size
        self.result_graph = nx.Graph()
        self.evaluated_end_nodes = set()

    def run(self):
        """
        Runs the ballistic line evaluation.
        """
        IDENTIFY_LOGGER.debug('Starting ballistic line evaluation.')
        # node = next(iter(self.graph.nodes))
        # self.find_loop(starting_node=node, initial_direction=None)
        while not nx.is_dominating_set(self.graph, self.result_graph.nodes):
            # First, try starting from an end node
            end_nodes = {
                node
                for node, deg in self.result_graph.degree if deg == 1
            }
            new_end_nodes = end_nodes - self.evaluated_end_nodes
            if new_end_nodes:
                node = new_end_nodes.pop()
                nbr = next(self.result_graph.neighbors(node))
                delta = self.coordinate_system.connecting_vector(
                    np.array(node), np.array(nbr)
                )
                # go in the opposite direction from the previous path
                initial_direction = -(delta / la.norm(delta))
                self.find_loop(
                    starting_node=node, initial_direction=initial_direction
                )
                self.evaluated_end_nodes.add(node)
            # Otherwise, pick an arbitrary starting node not in the vicinity of the result graph
            else:
                neighbor_graph = self.graph.edge_subgraph(
                    self.graph.edges(self.result_graph.nodes)
                )
                candidate_nodes = set(self.graph.nodes
                                      ) - set(neighbor_graph.nodes)
                node = candidate_nodes.pop()
                self.find_loop(starting_node=node, initial_direction=None)

    def find_loop(self, *, starting_node, initial_direction):
        """
        Find a loop from the given starting node, going (roughly) in the given
        direction. Stops when either the loop is closed (two nodes within the
        result graph found), or there are no more available steps.
        """
        IDENTIFY_LOGGER.debug(
            "Searching loop starting from node %s, with direction %s",
            starting_node, initial_direction
        )
        node = starting_node
        previous_direction = initial_direction
        self.result_graph.add_node(node)
        on_graph = False
        while True:
            new_node, new_direction = self.get_next_pos(
                node=node,
                previous_direction=previous_direction,
            )
            if new_node is None:
                self.evaluated_end_nodes.add(node)
                IDENTIFY_LOGGER.debug(
                    "Loop search finished -- no more nodes in given direction. Current node: %s, direction: %s",
                    node, previous_direction
                )
                assert previous_direction is not None
                break
            if new_node in self.result_graph:
                if on_graph:
                    IDENTIFY_LOGGER.debug(
                        "Loop search finished -- reached existing result nodes."
                    )
                    break
                on_graph = True
            else:
                on_graph = False
            self.result_graph.add_edge(node, new_node)
            node = new_node
            previous_direction = new_direction

    def get_next_pos(self, *, node, previous_direction):  # pylint: disable=too-many-locals
        """
        Determine the next node based on the current node and previous step
        direction.
        """
        neighbors_all = list(self.graph.neighbors(node))

        deltas_all = self.coordinate_system.connecting_vector(
            np.array(node), np.array(neighbors_all)
        )
        delta_norms_all = la.norm(deltas_all, axis=-1)

        neighbors = []
        unit_vecs = []
        for nbr, delta, delta_norm in zip(
            neighbors_all, deltas_all, delta_norms_all
        ):
            if delta_norm > 0:
                neighbors.append(nbr)
                unit_vecs.append(delta / delta_norm)
        unit_vecs = np.array(unit_vecs)

        if previous_direction is not None:
            direction_angles = unit_vecs @ previous_direction
        else:
            direction_angles = np.ones(len(unit_vecs))
        positive_angle_idx = np.where(direction_angles > 0)[0]
        if positive_angle_idx.size == 0:
            return None, None
        candidate_nodes = [neighbors[i] for i in positive_angle_idx]

        candidate_unit_vecs = unit_vecs[positive_angle_idx]
        angle_weight_matrix = candidate_unit_vecs @ candidate_unit_vecs.T

        weights = np.sum(np.maximum(angle_weight_matrix, 0), axis=-1)
        weights *= direction_angles[positive_angle_idx]

        # prefer nodes which are already in the result graph
        in_result_graph = np.array([
            nbr in self.result_graph.nodes for nbr in candidate_nodes
        ])
        weights *= 1 + _MULTIPLIER_EXISTING_NODE * in_result_graph.astype(int)

        node_degrees = dict(self.result_graph.degree)
        # prefer nodes which have a higher degree
        has_higher_degree = np.array([
            node_degrees.get(nbr, 0) > 2 for nbr in candidate_nodes
        ])
        weights *= 1 + _MULTIPLIER_HIGH_DEGREE * has_higher_degree.astype(int)
        # prefer nodes which have degree one
        has_low_degree = np.array([
            node_degrees.get(nbr, 0) == 1 for nbr in candidate_nodes
        ])
        weights *= 1 + _MULTIPLIER_LOW_DEGREE * has_low_degree.astype(int)

        # prefer nodes which are close to the current one
        distances_normalized = np.array([
            self.graph.edges[(node, neighbors[i])][_DISTANCE_KEY]
            for i in positive_angle_idx
        ]) / self.feature_size
        weights /= (1 + _MULTIPLIER_DISTANCE * distances_normalized)

        # avoid nodes which are much too close
        too_close = distances_normalized < 1e-3
        weights *= 1 - too_close.astype(int)

        idx = positive_angle_idx[np.argmax(weights)]
        chosen_candidate = neighbors[idx]

        return chosen_candidate, unit_vecs[idx]



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines functions for plotting the results of the identify step.
"""

from functools import singledispatch

import numpy as np
import scipy.linalg as la
from fsc.export import export

from .result import NodalPoint, NodalLine

from .._common_plot import _setup_plot


@export
def result(res, *, axis=None):
    """Plot the result of the identify step.

    Arguments
    ---------
    res : IdentificationResultContainer
        Result of the identify step.
    axis : matplotlib.axes.Axes, optional
        Axes on which the result is plotted.
    """
    fig, axis, _ = _setup_plot(res.coordinate_system.limits, axis=axis)
    feature_size = res.feature_size
    for identification_result in res:
        shape = identification_result.shape
        color = axis._get_lines.get_next_color()  # pylint: disable=protected-access
        if shape is None:
            _plot_positions(
                identification_result.positions, axis=axis, color=color
            )
        else:
            _plot_result(
                shape, axis=axis, color=color, feature_size=feature_size
            )
    return fig, axis


def _plot_positions(positions, *, axis, color):
    coordinates = list(np.array(list(positions)).T)
    axis.scatter(*coordinates, color=color)


@singledispatch
def _plot_result(shape, axis, color, feature_size):
    raise NotImplementedError


@export
@_plot_result.register(NodalPoint)
def nodal_point(shape, *, axis, color, feature_size=None):
    """
    Plot a nodal point.

    Arguments
    ---------
    shape : NodalPoint
        Nodal point to be plotted.
    axis : matplotlib.axes.Axes
        Axes on which to plot.
    color : str
        Color of the point.
    feature_size : float
        Distance between two nodal points at which they are considered distinct.
        This argument is not used in this function.
    """
    coordinates = [[val] for val in shape.position]
    axis.scatter(*coordinates, color=color)


@export
@_plot_result.register(NodalLine)
def nodal_line(shape, *, axis, color, feature_size=None):
    """
    Plot a nodal line.

    Arguments
    ---------
    shape : NodalLine
        Nodal line to be plotted.
    axis : matplotlib.axes.Axes
        Axes on which to plot.
    color : str
        Color of the nodal line.
    feature_size : float
        Distance between two nodal points at which they are considered distinct.
        Used for cutting the line when it goes across periodic boundaries.
    """
    if feature_size is None:
        feature_size = np.inf

    graph = shape.graph

    paths = _get_graph_paths(graph, feature_size=feature_size)

    if paths:
        for path in paths:
            axis.plot(*np.array(path).T, color=color)  # pylint: disable=not-an-iterable
    else:
        axis.scatter(*np.array(list(graph.nodes)).T, color=color)  # pylint: disable=not-an-iterable


def _get_graph_paths(graph, feature_size):
    """
    Separate a graph into paths, breaking when there is no neighbor or when
    passing across the periodic boundary.
    """
    working_graph = graph.copy()

    paths = []
    while working_graph.edges:
        curr_node = _get_next_starting_point(working_graph)
        curr_path = [curr_node]
        while True:
            try:
                next_node = next(working_graph.neighbors(curr_node))
            except StopIteration:
                paths.append(curr_path)
                break
            if la.norm(
                np.array(next_node) - np.array(curr_node)
            ) > 2 * feature_size:
                paths.append(curr_path)
                curr_path = [next_node]
            else:
                curr_path.append(next_node)

            working_graph.remove_edge(curr_node, next_node)
            curr_node = next_node
    return paths


def _get_next_starting_point(graph):
    nonzero_degree = [(node, degree) for node, degree in graph.degree
                      if degree > 0]
    return min(
        nonzero_degree,
        key=lambda val: val[1] if val[1] != 2 else float('inf')
    )[0]



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the container classes for object shapes.
"""

from types import SimpleNamespace

import numpy as np
import networkx as nx

from fsc.export import export
from fsc.hdf5_io import SimpleHDF5Mapping, subscribe_hdf5, to_hdf5, from_hdf5


@export
@subscribe_hdf5('nodefinder.nodal_point')
class NodalPoint(SimpleNamespace, SimpleHDF5Mapping):
    """
    Shape class defining a nodal point.

    Attributes
    ----------
    position : tuple(float)
        The position of the point.
    """
    HDF5_ATTRIBUTES = ['position']

    def __init__(self, position):
        self.position = position


@export
@subscribe_hdf5('nodefinder.nodal_line')
class NodalLine(SimpleNamespace):
    """
    Shape class defining a nodal line.

    Attributes
    ----------
    graph : networkx.Graph
        A graph describing the line.
    """
    def __init__(self, graph, degree_count):
        self.graph = graph
        self.degree_count = degree_count

    def __repr__(self):
        return 'NodalLine(graph=<{} nodes, {} edges>, degree_count={}, shape_name=\'{}\')'.format(
            len(self.graph.nodes), len(self.graph.edges), self.degree_count,
            self.shape_name
        )

    def to_hdf5(self, hdf5_handle):
        """
        Serialize the object and store in under the given HDF5 handle.
        """
        graph_group = hdf5_handle.create_group('graph')
        graph_group['nodes'] = np.array(list(self.graph.nodes))
        if self.graph.edges:
            graph_group['edges'] = np.array(list(self.graph.edges))

        degree_count_group = hdf5_handle.create_group('degree_count')
        to_hdf5(self.degree_count, degree_count_group)

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        """
        Derialize the object from the given HDF5 handle.
        """
        graph_group = hdf5_handle['graph']
        graph = nx.Graph()

        graph.add_nodes_from([tuple(n) for n in graph_group['nodes']])
        if 'edges' in graph_group:
            graph.add_edges_from([(tuple(p1), tuple(p2))
                                  for p1, p2 in graph_group['edges']])

        degree_count = from_hdf5(hdf5_handle['degree_count'])
        return cls(graph=graph, degree_count=degree_count)

    @property
    def shape_name(self):
        """
        Describes the shape of the line, as inferred from the degree count.
        """
        shape_lookup = {
            tuple(): 'CLOSED LOOP',
            ((1, 2), ): 'OPEN LINE',
        }
        return shape_lookup.get(
            tuple(sorted(self.degree_count.items())), 'UNKNOWN'
        )



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the Controller, which implements the evaluation of the search step.
"""

import os
import numbers
import asyncio
import tempfile
from collections import ChainMap

import numpy as np
from fsc.export import export
from fsc.async_tools import PeriodicTask, wrap_to_coroutine

from .. import io
from ..coordinate_system import CoordinateSystem
from .result import SearchResultContainer, ControllerState
from ._queue import SimplexQueue, PositionQueue
from ._minimization import run_minimization
from ._fake_potential import FakePotential
from ._logging import SEARCH_LOGGER
from ._mesh_helper import _generate_mesh_simplices
from .refinement_stencil import get_auto_stencil

_DIST_CUTOFF_FACTOR = 3


@export
class Controller:
    """
    Implementation class for the :func:`.search.run` function.

    Arguments are the same as defined in :func:`.search.run`.
    """
    def __init__(
        self,
        *,
        gap_fct,
        limits,
        periodic,
        initial_state,
        save_file,
        save_delay=5.,
        load,
        load_quiet,
        initial_mesh_size,
        force_initial_mesh,
        gap_threshold,
        feature_size,
        nelder_mead_kwargs,
        num_minimize_parallel,
        refinement_stencil,
        use_fake_potential=True,
        recheck_pos_dist=True,
        recheck_count_cutoff=0,
        simplex_check_cutoff=0
    ):
        self.gap_fct = wrap_to_coroutine(gap_fct)

        self.coordinate_system = CoordinateSystem(
            limits=limits, periodic=periodic
        )
        self.dim, initial_mesh_size = self.check_dimensions(
            limits, initial_mesh_size
        )
        self.save_file = save_file
        self.save_delay = save_delay

        self.dist_cutoff = feature_size / _DIST_CUTOFF_FACTOR
        self.state = self.create_state(
            initial_state=initial_state,
            load=load,
            load_quiet=load_quiet,
            initial_mesh_size=initial_mesh_size,
            force_initial_mesh=force_initial_mesh,
            gap_threshold=gap_threshold,
            dist_cutoff=self.dist_cutoff
        )
        if use_fake_potential:
            self.fake_potential = FakePotential(
                result=self.state.result,
                width=self.dist_cutoff,
            )
        else:
            self.fake_potential = None

        if isinstance(
            refinement_stencil, str
        ) and refinement_stencil == 'auto':
            refinement_stencil = get_auto_stencil(dim=self.dim)
        if refinement_stencil is not None:
            self.refinement_stencil = refinement_stencil * self.dist_cutoff
        else:
            self.refinement_stencil = None
        self.num_minimize_parallel = num_minimize_parallel
        self.nelder_mead_kwargs = ChainMap(
            nelder_mead_kwargs, {
                'ftol': 0.05 * gap_threshold,
                'xtol': 0.03 * self.dist_cutoff
            }
        )

        self.task_futures = set()
        self.recheck_pos_dist = recheck_pos_dist
        self.recheck_count_cutoff = recheck_count_cutoff
        self.simplex_check_cutoff = simplex_check_cutoff

    @staticmethod
    def check_dimensions(limits, mesh_size):
        """
        Check that the dimensions of the given inputs match.
        """
        if isinstance(mesh_size, numbers.Integral):
            mesh_size = tuple(mesh_size for _ in range(len(limits)))
        dim_limits = len(limits)
        dim_mesh_size = len(mesh_size)
        if not dim_limits == dim_mesh_size:
            raise ValueError(
                'Inconsistent dimensions given: limits: {}, mesh_size: {}'.
                format(dim_limits, dim_mesh_size)
            )
        return dim_limits, mesh_size

    def create_state(
        self, *, initial_state, load, load_quiet, initial_mesh_size,
        force_initial_mesh, gap_threshold, dist_cutoff
    ):
        """
        Load or create the initial state of the calculation.
        """
        if load:
            if initial_state is not None:
                raise ValueError(
                    "Cannot set the initial state explicitly and setting 'load=True' simultaneously."
                )
            try:
                initial_state = io.load(self.save_file)
            except IOError as exc:
                if not load_quiet:
                    raise exc
        if initial_state is not None:
            result = SearchResultContainer(
                coordinate_system=self.coordinate_system,
                minimization_results=initial_state.result.minimization_results,
                gap_threshold=gap_threshold,
                dist_cutoff=dist_cutoff,
                refined_results=initial_state.result.refined_results
            )
            simplex_queue = SimplexQueue(
                objects=initial_state.simplex_queue.objects
            )
            position_queue = PositionQueue(
                objects=initial_state.position_queue.objects
            )
            if force_initial_mesh:
                simplex_queue.add_objects(
                    self.get_initial_simplices(
                        initial_mesh_size=initial_mesh_size
                    )
                )
        else:
            result = SearchResultContainer(
                coordinate_system=self.coordinate_system,
                gap_threshold=gap_threshold,
                dist_cutoff=dist_cutoff,
            )
            simplex_queue = SimplexQueue(
                self.get_initial_simplices(initial_mesh_size)
            )
            position_queue = PositionQueue()
        return ControllerState(
            result=result,
            simplex_queue=simplex_queue,
            position_queue=position_queue
        )

    def get_initial_simplices(self, initial_mesh_size):
        return _generate_mesh_simplices(
            limits=self.coordinate_system.limits,
            mesh_size=initial_mesh_size,
            periodic=self.coordinate_system.periodic
        )

    async def run(self):
        await self.create_tasks()

    async def create_tasks(self):
        """
        Create minimization tasks until the calculation is finished.
        """
        async with PeriodicTask(self.save, delay=self.save_delay):
            while (
                not self.state.simplex_queue.finished
            ) or self.state.position_queue.has_queued:
                # if (not self.state.simplex_queue.has_queued) and (self.state.position_queue.has_queued):
                while (
                    self.state.simplex_queue.num_running <
                    self.num_minimize_parallel
                ):
                    while not self.state.simplex_queue.has_queued:
                        if self.state.position_queue.has_queued:
                            pos = self.state.position_queue.pop_queued()
                            if (not self.recheck_pos_dist
                                ) or self._check_pos_refinement(
                                    pos,
                                    count_cutoff=self.recheck_count_cutoff
                                ):
                                self.state.simplex_queue.add_objects(
                                    pos + self.refinement_stencil
                                )
                                self.state.result.set_refined(np.array(pos))
                            else:
                                SEARCH_LOGGER.debug(
                                    'Discarding refinement of position {}'.
                                    format(pos)
                                )
                        else:
                            break
                    if self.state.simplex_queue.has_queued:
                        simplex = self.state.simplex_queue.pop_queued()
                        if self._check_simplex(simplex):
                            self.schedule_minimization(simplex)
                        else:
                            self.state.simplex_queue.set_finished(simplex)
                    else:
                        break

                await asyncio.sleep(0.)

                # Retrieve all exceptions, to avoid 'exception never retrieved'
                # warning, but raise only the first one.
                done_futures = [fut for fut in self.task_futures if fut.done()]
                exceptions = [fut.exception() for fut in done_futures]
                exceptions = [exc for exc in exceptions if exc is not None]
                if exceptions:
                    raise exceptions[0]

                self.task_futures.difference_update(done_futures)
        await asyncio.gather(*self.task_futures)

    def schedule_minimization(self, simplex):
        SEARCH_LOGGER.debug(
            'Scheduling minimization of simplex {}'.format(simplex)
        )
        self.task_futures.add(asyncio.ensure_future(self.run_simplex(simplex)))

    async def run_simplex(self, simplex):
        """
        Run the minimization for a given starting simplex.
        """
        result = await run_minimization(
            self.gap_fct,
            initial_simplex=simplex,
            fake_potential=self.fake_potential,
            nelder_mead_kwargs=self.nelder_mead_kwargs,
        )
        self.process_result(result)
        self.state.simplex_queue.set_finished(simplex)

    def process_result(self, result):
        """
        Update the state with a given result, and add new simplices if needed.
        """
        is_node = self.state.result.add_result(result)
        if is_node and self.refinement_stencil is not None:
            pos = result.pos
            SEARCH_LOGGER.info('Found node at position {}'.format(pos))
            if self._check_pos_refinement(pos):
                SEARCH_LOGGER.info('Scheduling refinement around node.')
                self.state.position_queue.add_objects([pos])

    def _check_pos_refinement(self, pos, count_cutoff=0):
        """
        Check whether a given position should be scheduled for refinement. An
        optional cutoff for the number of positions which are allowed to be
        within the cutoff distance can be given.
        """
        count = 0
        for dist in self.state.result.get_refined_neighbour_distance_iterator(
            pos
        ):
            if dist < self.dist_cutoff:
                count += 1
            if count > count_cutoff:
                return False
        return True

    def _check_simplex(self, simplex):
        """
        Check if a simplex should be evaluated. Returns False if all vertices
        of the simplex are within dist_cutoff from an existing node.
        """
        count = 0
        for pos in simplex:
            for dist in self.state.result.get_neighbour_distance_iterator(pos):
                if dist < self.dist_cutoff:
                    count += 1
                    if count > self.simplex_check_cutoff:
                        return False
                    break
        return True

    def save(self):
        """
        Store the current ControllerState to the save file.
        """
        if self.save_file and self.state.needs_saving:
            with tempfile.NamedTemporaryFile(
                dir=os.path.dirname(self.save_file), delete=False
            ) as tmpf:
                try:
                    io.save(self.state, tmpf.name)
                    os.rename(tmpf.name, self.save_file)
                    self.state.needs_saving = False
                except Exception as exc:
                    os.remove(tmpf.name)
                    raise exc



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines helper functions for plotting.
"""


def _setup_plot(limits, axis=None):
    """Sets up the figure axes and handles options common to all plots."""
    # import is here s.t. the import of the package does not fail
    # if matplotlib is not present
    import matplotlib.pyplot as plt  # pylint: disable=import-outside-toplevel

    dim = len(limits)
    if dim == 3:
        is_3d = True
        from mpl_toolkits.mplot3d import Axes3D  # pylint: disable=import-outside-toplevel,unused-import
    else:
        is_3d = False
        assert dim == 2
    # create axis if it does not exist
    if axis is None:
        fig = plt.figure()
        axis = fig.add_subplot(111, projection='3d' if is_3d else None)
    else:
        fig = None

    axis.set_xlim(*limits[0])
    axis.set_ylim(*limits[1])
    if is_3d:
        axis.set_zlim(*limits[2])

    return fig, axis, is_3d



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the functions used to evaluate the shape of a given cluster of points.
"""

from ._run import *

__all__ = _run.__all__  # pylint: disable=undefined-variable



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Submodule for searching the nodal features given a gap function. The result is
given as a "point-cloud" of nodes.
"""

from ._run import *
from . import result
from . import plot

__all__ = _run.__all__ + ['result', 'plot']  # pylint: disable=undefined-variable



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the clustering function used to group nodal points.
"""

import numpy as np
import networkx as nx
from fsc.export import export

_DISTANCE_KEY = '_distance'


@export
def create_clusters(positions, *, feature_size, coordinate_system):
    """
    Create clusters from the given positions, by joining all positions which are
    less than twice the ``feature_size`` apart.

    Arguments
    ---------
    positions : list(list(float))
        The list of positions to cluster.
    feature_size : float
        Distance between two nodes where they are considered to belong to different clusters.
    coordinate_system : CoordinateSystem
        Coordinate system used to calculate distances between positions.

    Returns
    -------
    list(nx.Graph) :
        A list of connected graphs, each representing one cluster.
    """
    graph = _create_graph(
        positions,
        feature_size=feature_size,
        coordinate_system=coordinate_system
    )
    return [
        nx.freeze(graph.subgraph(nodes).copy())
        for nodes in nx.algorithms.connected_components(graph)
    ]


def _create_graph(positions, *, feature_size, coordinate_system):
    """
    Create a graph from the positions, where nodes which are less than the feature
    size apart are connected by an edge, whose 'distance' weight is their
    distance.
    """
    graph = nx.Graph()

    pos_unique = list(set(tuple(pos) for pos in positions))
    graph.add_nodes_from(pos_unique)
    pos_unique_array = np.array(pos_unique)
    for i, (pos, pos_arr) in enumerate(zip(pos_unique[:-1], pos_unique_array)):
        offset = i + 1
        candidates = pos_unique_array[offset:]
        distances = coordinate_system.distance(pos_arr, candidates)
        for idx in np.flatnonzero(distances <= feature_size):
            nbr = pos_unique[idx + offset]
            dist = distances[idx]
            graph.add_edge(pos, nbr, **{_DISTANCE_KEY: dist})
    return graph



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Implements line evaluation with the shortest path method.
"""

import networkx as nx

from ..._cluster import _DISTANCE_KEY
from ..._logging import IDENTIFY_LOGGER

_WEIGHT_KEY = '_weight'
_MAX_NUM_PATHS = 20

def _evaluate_line_shortest_path(  # pylint: disable=too-many-locals,too-many-statements,too-many-branches
    *,
    graph,
    coordinate_system,  # pylint: disable=unused-argument
    feature_size
):
    """
    Evaluate the positions of a nodal line using the 'shortest path' method.
    """
    work_graph = graph.copy()
    distance_func = lambda x: x + x**2
    for edge in work_graph.edges:
        edge_attrs = work_graph.edges[edge]
        dist = edge_attrs[_DISTANCE_KEY]
        edge_attrs[_WEIGHT_KEY] = feature_size * distance_func(
            dist / feature_size
        )

    candidate_positions = set(work_graph.nodes)

    start_positions = set()
    first_pos = candidate_positions.pop()
    start_positions.add(first_pos)
    candidate_positions -= set(work_graph.neighbors(first_pos))

    result_graph = nx.Graph()
    result_graph.add_node(first_pos)

    high_degree_nodes = set()
    degree_one_nodes = set()

    while candidate_positions:

        end_pos = candidate_positions.pop()
        end_neighbours = set(work_graph.neighbors(end_pos)) | {end_pos}
        candidate_positions -= end_neighbours

        for start_pos in start_positions:
            # prioritize "crossings"
            _update_high_degree_nodes(
                work_graph, result_graph, high_degree_nodes
            )
            # prioritize ends
            _update_degree_one_nodes(
                work_graph, result_graph, degree_one_nodes
            )
            tmp_graph = work_graph.copy()

            start_neighbours = set(work_graph.neighbors(start_pos)
                                   ) | {start_pos}
            start_end_neighbours = start_neighbours | end_neighbours

            unweighted_path_length = nx.algorithms.shortest_path_length(
                tmp_graph,
                source=start_pos,
                target=end_pos,
            )
            if unweighted_path_length <= 4:
                single_path_only = True
                IDENTIFY_LOGGER.debug(
                    'Positions %s and %s are too close, calculating only one path.',
                    start_pos, end_pos
                )
            else:
                single_path_only = False
            for num_paths in range(_MAX_NUM_PATHS):
                if single_path_only and num_paths == 1:
                    break
                try:
                    path = nx.algorithms.shortest_path(
                        tmp_graph,
                        source=start_pos,
                        target=end_pos,
                        weight=_WEIGHT_KEY,
                    )
                except nx.NetworkXNoPath:
                    break

                if num_paths > 3:
                    print(path)
                new_neighbours = set()
                for node in path:
                    new_neighbours.update(tmp_graph.neighbors(node))
                    new_neighbours.add(node)
                candidate_positions -= new_neighbours

                nodes_to_remove = new_neighbours - start_end_neighbours
                assert nodes_to_remove or single_path_only

                edges = list(zip(path, path[1:]))
                result_graph.add_edges_from(edges)
                for edge in edges:
                    work_graph.edges[edge][_WEIGHT_KEY] = 0
                    tmp_graph.edges[edge][_WEIGHT_KEY] = 0

                tmp_graph.remove_nodes_from(nodes_to_remove)

            if num_paths > 0:
                IDENTIFY_LOGGER.debug(
                    'Found {} path(s) from point {} to point {}.'.format(
                        num_paths, start_pos, end_pos
                    )
                )
            else:
                IDENTIFY_LOGGER.warning(
                    'No paths to point {} found.'.format(end_pos)
                )
        start_positions.add(end_pos)

    return result_graph


def _update_high_degree_nodes(graph, result_graph, high_degree_nodes):
    """
    Update the high degree nodes, and adjust the weight of the edges for new
    high degree nodes.
    """
    for node, deg in result_graph.degree:
        if deg > 2 and node not in high_degree_nodes:
            high_degree_nodes.add(node)
            _update_neighbour_weights(node, graph=graph, multiplier=0.5)


def _update_degree_one_nodes(graph, result_graph, degree_one_nodes):
    """
    Update the nodes of degree one. Decrease the weight of edges for
    new nodes of degree one, and increase the weight for those whose
    degree is no longer one.
    """
    current_deg_one_nodes = set(
        node for node, deg in result_graph.degree if deg == 1
    )
    new_nodes = current_deg_one_nodes - degree_one_nodes
    outdated_nodes = degree_one_nodes - current_deg_one_nodes
    for node in new_nodes:
        _update_neighbour_weights(node, graph=graph, multiplier=0.1)
    for node in outdated_nodes:
        _update_neighbour_weights(node, graph=graph, multiplier=10)
    degree_one_nodes.clear()
    degree_one_nodes.update(current_deg_one_nodes)


def _update_neighbour_weights(node, graph, multiplier):
    for nbr in graph.neighbors(node):
        graph.edges[(node, nbr)][_WEIGHT_KEY] *= multiplier



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the logger for the search submodule.
"""

import logging

SEARCH_LOGGER = logging.getLogger('nodefinder.search')



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the logger for the main module.
"""

import sys
import logging

MAIN_LOGGER = logging.getLogger('nodefinder')
DEFAULT_HANDLER = logging.StreamHandler(sys.stdout)
FORMATTER = logging.Formatter(
    fmt='%(asctime)s - %(levelname)s - %(message)s',
)
DEFAULT_HANDLER.setFormatter(FORMATTER)
MAIN_LOGGER.addHandler(DEFAULT_HANDLER)



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the function running the minimization, including the fake potential.
"""

from types import MappingProxyType

from fsc.export import export

from ..result._minimization import JoinedMinimizationResult
from ._nelder_mead import root_nelder_mead


def add_fake_potential(fake_pot, func):
    async def evaluate(x):
        return (await func(x)) + fake_pot(x)

    return evaluate


@export
async def run_minimization(
    func,
    *,
    initial_simplex,
    fake_potential=None,
    nelder_mead_kwargs=MappingProxyType({})
):
    """Runs the minimization, including handling the fake potential.

    Runs the minimization for a given function and initial simplex. If a
    fake potential is given, the Nelder-Mead algorithm is first run with the
    fake potential, and then continued without the fake potential.
    The final simplex of the minimization with fake potential is enlarged and
    used as initial simplex for the second Nelder-Mead run.

    Arguments
    ---------
    func : collections.abc.Callable
        Function or coroutine describing the potential to be minimized.
    initial_simplex : numpy.ndarray
        Coordinates of the initial simplex.
    fake_potential : collections.abc.Callable
        Function describing the fake potential.
    nelder_mead_kwargs : collections.abc.Mapping
        Keyword arguments passed to the Nelder-Mead algorithm.
    """
    if fake_potential is not None:
        # TODO: Check if deepcopying fake potential is valid / better.
        # possible issue when not deep-copying is that the fake potential may
        # change during minimization, and the Nelder-Mead algorithm could get
        # horribly stuck when the current best value is within the 'infinite'
        # region.
        modified_kwargs = dict(nelder_mead_kwargs)
        modified_kwargs['ftol'] = float('inf')
        res_fake = await root_nelder_mead(
            func=add_fake_potential(fake_potential, func),
            initial_simplex=initial_simplex,
            **modified_kwargs
        )
        simplex_final = res_fake.simplex_history[-1]
        simplex_blowup = simplex_final[
            0] + 1.5 * (simplex_final - simplex_final[0])

        res = await root_nelder_mead(
            func=func, initial_simplex=simplex_blowup, **nelder_mead_kwargs
        )

        return JoinedMinimizationResult(child=res, ancestor=res_fake)
    else:
        return await root_nelder_mead(
            func=func, initial_simplex=initial_simplex, **nelder_mead_kwargs
        )



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the data class for the controller state.
"""

from fsc.export import export
from fsc.hdf5_io import SimpleHDF5Mapping, subscribe_hdf5


@export
@subscribe_hdf5('nodefinder.controller_state')
class ControllerState(SimpleHDF5Mapping):
    """
    Container class for the current result and queue of the :func:`.search.run`
    function.
    """
    HDF5_ATTRIBUTES = ['result', 'simplex_queue', 'position_queue']

    def __init__(self, *, result, simplex_queue, position_queue):
        self.result = result
        self.simplex_queue = simplex_queue
        self.position_queue = position_queue

    @property
    def needs_saving(self):
        return self.result.needs_saving or self.simplex_queue.needs_saving or self.position_queue.needs_saving

    @needs_saving.setter
    def needs_saving(self, value):
        assert not value
        self.result.needs_saving = value
        self.simplex_queue.needs_saving = value
        self.position_queue.needs_saving = value



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the dispatcher function for evaluating a line, delegating to the specific
method implementation.
"""

from collections import Counter

from ...result import NodalLine

from ._shortest_path import _evaluate_line_shortest_path
from ._dominating_set import _evaluate_line_dominating_set
from ._ballistic import _evaluate_line_ballistic

_METHOD_FUNC_LOOKUP = dict(
    shortest_path=_evaluate_line_shortest_path,
    dominating_set=_evaluate_line_dominating_set,
    ballistic=_evaluate_line_ballistic
)


def _evaluate_line(graph, coordinate_system, feature_size, method='ballistic'):
    """
    Evaluate the positions of a nodal line.
    """

    try:
        method_func = _METHOD_FUNC_LOOKUP[method]
    except KeyError:
        raise ValueError("Invalid value for 'method': {}".format(method))
    result_graph = method_func(
        graph=graph,
        coordinate_system=coordinate_system,
        feature_size=feature_size
    )
    return NodalLine(
        graph=result_graph, degree_count=_create_degree_count(result_graph)
    )


def _create_degree_count(graph):
    degree_counter = Counter([val for pos, val in graph.degree])
    degree_counter.pop(2, None)
    return degree_counter



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines free functions to serialize / deserialize nodefinder objects to HDF5.

These are aliases for the functions defined in :mod:`fsc.hdf5_io`, meaning that
they can also handle other objects which are registered with the same system.
"""

from fsc.hdf5_io import save, load

__all__ = ['save', 'load']



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the function which runs the identify step.
"""

from fsc.export import export

from ..search.result import ControllerState
from ..search._controller import _DIST_CUTOFF_FACTOR

from .result import IdentificationResult, IdentificationResultContainer
from ._cluster import create_clusters
from ._dimension import calculate_dimension
from ._evaluate import evaluate_cluster
from ._logging import IDENTIFY_LOGGER


@export
def run(result, feature_size=None, evaluate_line_method='ballistic'):
    """Identify the nodal clusters from a :func:`.search.run` result.

    Arguments
    ---------
    result : SearchResultContainer
        Result of the search step.
    feature_size : float, optional
        Distance between two nodal points at which they are considered distinct.
        Uses the ``feature_size`` used in the search step by default.
    """
    if feature_size is None:
        feature_size = result.dist_cutoff * _DIST_CUTOFF_FACTOR
    if isinstance(result, ControllerState):
        result = result.result
    positions = [node.pos for node in result.nodes]
    coordinate_system = result.coordinate_system
    return run_from_positions(
        positions=positions,
        coordinate_system=coordinate_system,
        feature_size=feature_size,
        evaluate_line_method=evaluate_line_method,
    )


@export
def run_from_positions(
    positions,
    *,
    coordinate_system,
    feature_size,
    evaluate_line_method='shortest_path'
):
    """Identify the nodal clusters from a list of positions.

    Arguments
    ---------
    positions : list(tuple(float))
        The list of nodal positions.
    coordinate_system : CoordinateSystem
        Coordinate system used to calculate distances.
    feature_size : float
        Distance between two nodal points at which they are considered distinct.
    """
    IDENTIFY_LOGGER.debug('Calculating clusters.')
    clusters = create_clusters(
        positions,
        coordinate_system=coordinate_system,
        feature_size=feature_size
    )
    results = []
    for graph in clusters:
        IDENTIFY_LOGGER.debug('Calculating cluster dimension.')
        dim = calculate_dimension(
            graph=graph,
            coordinate_system=coordinate_system,
            feature_size=feature_size
        )
        IDENTIFY_LOGGER.debug(
            'Evaluating result with dimension {}.'.format(dim)
        )
        res = IdentificationResult(
            positions=list(graph.nodes),
            dimension=dim,
            shape=evaluate_cluster(
                graph=graph,
                dim=dim,
                coordinate_system=coordinate_system,
                feature_size=feature_size,
                evaluate_line_method=evaluate_line_method
            )
        )
        results.append(res)
    return IdentificationResultContainer(
        coordinate_system=coordinate_system,
        feature_size=feature_size,
        results=results
    )



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Submodule for identifying nodal features such as points and lines from a
point-cloud of nodes as created by the :mod:`.search` submodule.
"""

from ._run import *
from . import result
from . import plot

__all__ = ['result', 'plot'] + _run.__all__  # pylint: disable=undefined-variable



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the coordinate system class.
"""

import numpy as np
import scipy.linalg as la
from fsc.export import export
from fsc.hdf5_io import SimpleHDF5Mapping, subscribe_hdf5


@export
@subscribe_hdf5('nodefinder.coordinate_system')
class CoordinateSystem(SimpleHDF5Mapping):
    """
    Defines a "box" coordinate system, which is used to calculate the distances
    between points, and map between fractional and real coordinates.

    Attributes
    ----------
    limits : numpy.ndarray
        Limits of the coordinates, given as a the minimum and
        maximum value for each dimension.
    periodic : bool
        Determies if periodic boundary conditions are used.
    dim : int
        Dimension of the coordinate system.
    size : numpy.ndarray
        Size of the coordinate system in each dimension.
    """
    HDF5_ATTRIBUTES = ['limits', 'periodic']

    def __init__(self, *, limits, periodic=True):
        self.limits = np.array([sorted(x) for x in limits])
        self.periodic = periodic
        self.dim = len(self.limits)
        self.size = self._upper_limits - self._lower_limits

    def __repr__(self):
        return 'CoordinateSystem(limits={0.limits!r}, periodic={0.periodic!r})'.format(
            self
        )

    @property
    def _lower_limits(self):
        return self.limits[:, 0]

    @property
    def _upper_limits(self):
        return self.limits[:, 1]

    def get_frac(self, pos):
        """
        Get the fractional coordinates from the real position.
        """
        frac = (pos - self._lower_limits) / self.size
        if self.periodic:
            frac %= 1
        return frac

    def get_pos(self, frac):
        """
        Get the real position from the fractional coordinates.
        """
        return (frac * self.size) + self._lower_limits

    def distance(self, pos1, pos2):
        """
        Get the distance between two positions.
        """
        if self.periodic:
            delta = (pos2 - pos1) % self.size
            delta = np.minimum(self.size - delta, delta)  # pylint: disable=assignment-from-no-return,useless-suppression
        else:
            delta = pos2 - pos1
        return la.norm(delta, axis=-1)

    def connecting_vector(self, pos1, pos2):
        """
        Get the shortest vector connecting two positions.
        """
        if not self.periodic:
            return pos2 - pos1
        else:
            delta = (pos2 - pos1) % self.size
            delta_negative = delta - self.size
            res = np.where(-delta_negative < delta, delta_negative, delta)
            assert np.all(self.distance((res + pos1), pos2) < 1e-8)
            return res

    def average(self, positions):
        """
        Get the average position from a list of positions.
        """
        if not self.periodic:
            return np.average(positions, axis=0)
        else:
            origin = positions[0]
            deltas = [self.connecting_vector(origin, pos) for pos in positions]
            return (origin + np.average(deltas, axis=0)) % self.size

    def normalize_position(self, pos):
        """
        Normalize the position by mapping it into the limits for periodic
        boundary conditions for the periodic case.
        """
        if self.periodic:
            return ((pos - self._lower_limits) %
                    self.size) + self._lower_limits
        else:
            return pos



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""A tool to find and identify nodal features in band structures.
"""

__version__ = '0.1.1'

from . import coordinate_system
from . import search
from . import identify
from . import io
from . import _logging

__all__ = ['search', 'identify', 'io', 'coordinate_system']



# ******NOTICE***************
# optimize.py module by Travis E. Oliphant
#
# You may copy and use this module as you see fit with no
# guarantee implied provided you keep this notice in all copies.
# *****END NOTICE************
#
# The additional license terms given in ADDITIONAL_TERMS.txt apply to this
# file.

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the result classes for the minimization.
"""

from types import SimpleNamespace

import numpy as np
from fsc.export import export
from fsc.hdf5_io import subscribe_hdf5, SimpleHDF5Mapping, HDF5Enabled


@export
@subscribe_hdf5(
    'nodefinder.joined_minimization_result',
    extra_tags=['nodefinder.joined_result']
)
class JoinedMinimizationResult(SimpleHDF5Mapping):
    """
    Wrapper for minimization results that result from two steps.

    Attributes
    ----------
    ancestor : MinimizationResult
        Result of the first minimization run.
    child : MinimizationResult
        Result of the second minimization run.
    """
    JOIN_KEYS = [
        'num_fev', 'num_iter', 'simplex_history', 'fun_simplex_history'
    ]
    HDF5_ATTRIBUTES = ['ancestor', 'child']

    def __init__(self, *, child, ancestor):
        self.child = child
        self.ancestor = ancestor

    def __getattr__(self, key):
        """
        Joins child and ancestor values where that makes sense, and returns the
        child value otherwise.
        """
        if key in self.JOIN_KEYS:
            return self._join(
                getattr(self.ancestor, key), getattr(self.child, key)
            )
        else:
            return getattr(self.child, key)

    @staticmethod
    def _join(obj1, obj2):
        """
        Helper function to join two result values.
        """
        if isinstance(obj1, np.ndarray):
            return np.concatenate([obj1, obj2])
        else:
            return obj1 + obj2


@export
@subscribe_hdf5('nodefinder.minimization_result')
class MinimizationResult(SimpleNamespace, HDF5Enabled):
    """ Represents the optimization result.

    Attributes
    ----------
    pos : ndarray
        The solution of the optimization.
    value : ndarray
        Value of objective function.
    success : bool
        Whether or not the optimizer exited successfully.
    status : int
        Termination status of the optimizer. Its value depends on the
        underlying solver. Refer to `message` for details.
    message : str
        Description of the cause of the termination.
    num_fev : int
        Number of evaluations of the objective functions.
    num_iter : int
        Number of iterations performed by the optimizer.
    simplex_history : ndarray, optional
        History of simplex values.
    fun_simplex_history : ndarray, optional
        History of function values of the simplex.
    """
    def to_hdf5(self, hdf5_handle):
        for key, val in self.__dict__.items():
            assert key != 'type_tag'
            hdf5_handle[key] = val

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return cls(
            **{
                key: val[()]
                for key, val in hdf5_handle.items() if key != 'type_tag'
            }
        )



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Define the result classes for the identification step.
"""

from ._containers import *
from ._shapes import *

__all__ = _containers.__all__ + _shapes.__all__  # pylint: disable=undefined-variable



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the container classes for the identification results.
"""

from types import SimpleNamespace

import numpy as np
from fsc.export import export
from fsc.hdf5_io import subscribe_hdf5, SimpleHDF5Mapping, HDF5Enabled, to_hdf5, from_hdf5


@export
@subscribe_hdf5('nodefinder.identification_result_container')
class IdentificationResultContainer(SimpleNamespace, SimpleHDF5Mapping):
    """Container class for the result of the identification step.

    Attributes
    ----------
    coordinate_system : CoordinateSystem
        The coordinate system of the problem.
    results : list(IdentificationResult)
        List of identified objects.
    feature_size : float
        The ``feature_size`` used when identifying the objects.
    """
    HDF5_ATTRIBUTES = ['coordinate_system', 'results', 'feature_size']

    def __init__(self, *, coordinate_system, feature_size, results=()):
        self.coordinate_system = coordinate_system
        self.results = results
        self.feature_size = feature_size

    def __iter__(self):
        return iter(self.results)

    def __getitem__(self, idx):
        return self.results[idx]

    def __len__(self):
        return len(self.results)


@export
@subscribe_hdf5('nodefinder.identification_result')
class IdentificationResult(SimpleNamespace, HDF5Enabled):
    """Contains the attributes of an identified object.

    Attributes
    ----------
    positions : list(tuple(float))
        Positions of the nodal points making up the object.
    shape : :obj:`None` or NodalPoint or NodalLine
        Shape of the identified object. If the shape could not be identified, it
        is set to ``None``.
    dimension : int
        Dimension of the identified object. Is set to ``None`` if the dimension
        is ambiguous.
    """

    # HDF5_ATTRIBUTES = ['positions', 'shape', 'dimension']

    def __init__(self, positions, dimension, shape=None):
        self.positions = [tuple(pos) for pos in positions]
        self.dimension = dimension
        self.shape = shape

    def __repr__(self):
        return 'IdentificationResult(dimension={}, shape={}, positions=<{} values>)'.format(
            self.dimension, self.shape, len(self.positions)
        )

    def to_hdf5(self, hdf5_handle):
        to_hdf5(self.dimension, hdf5_handle.create_group('dimension'))
        hdf5_handle['positions'] = np.array(self.positions)
        to_hdf5(self.shape, hdf5_handle.create_group('shape'))

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        shape = from_hdf5(hdf5_handle['shape'])
        try:
            dimension = hdf5_handle['dimension'][()]
        except AttributeError:
            dimension = from_hdf5(hdf5_handle['dimension'])
        try:
            positions = [tuple(x) for x in hdf5_handle['positions'][()]]
        except AttributeError:
            positions = from_hdf5(hdf5_handle['positions'])
        return cls(positions=positions, dimension=dimension, shape=shape)



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines functions for plotting the results of the search step.
"""

import itertools
from types import MappingProxyType

import numpy as np
from fsc.export import export

from .._common_plot import _setup_plot


@export
def points(result, *, axis=None):
    """
    Plot the nodal points of a given result.

    Arguments
    ---------
    result : SearchResultContainer
        Result whose nodal points should be plotted.
    axis : matplotlib.axes.Axes, optional
        Axes on which the points should be plotted.
    """
    fig, axis, is_3d = _setup_plot(result.coordinate_system.limits, axis=axis)
    x_coords = []
    y_coords = []
    if is_3d:
        z_coords = []
    vals = []
    for node in result.nodes:
        pos = node.pos
        x, y, *z = pos
        x_coords.append(x)
        y_coords.append(y)
        if is_3d:
            z_coords.append(z[0])
        vals.append(node.value)

    if is_3d:
        coords = [x_coords, y_coords, z_coords]
    else:
        coords = [x_coords, y_coords]

    axis.scatter(*coords, c=vals)
    return fig, axis


@export
def simplices(
    result,
    *,
    nodes=(),
    axis=None,
    line_settings=MappingProxyType(dict(color='C0'))
):
    """
    Plot the simplices used in the minimization for a given node.

    Arguments
    ---------
    result : SearchResultContainer
        Result of the search step.
    nodes : list(MinimizationResult)
        Nodes for which the simplex history should be plotted.
    axis : matplotlib.axes.Axes, optional
        Axes on which the points should be plotted.
    line_settings : dict
        Keyword arguments passed to :meth:`matplotlib.axes.Axes.plot`.
    """
    fig, axis, _ = _setup_plot(result.coordinate_system.limits, axis=axis)
    for node in nodes:
        for simplex in node.simplex_history:
            _plot_simplex(axis=axis, simplex=simplex, **line_settings)
    return fig, axis


@export
def stencil(
    stcl,
    *,
    axis=None,
    line_settings=MappingProxyType(dict(color='C0')),
    origin_settings=MappingProxyType(dict(color='C1')),
    plot_sphere=True,
    sphere_settings=MappingProxyType(dict(color='C2', alpha=0.2))
):
    """
    Plot a refinement stencil.
    """
    dim = stcl.shape[-1]
    vertices = stcl.reshape(-1, dim)
    limits = list(zip(np.min(vertices, axis=0), np.max(vertices, axis=0)))
    fig, axis, _ = _setup_plot(limits, axis=axis)
    for simplex in stcl:
        _plot_simplex(axis=axis, simplex=simplex, **line_settings)
    if origin_settings is not None:
        axis.scatter(*np.zeros(dim), **origin_settings)
    if plot_sphere:
        if dim == 2:
            phi = np.linspace(0, 2 * np.pi, 200)
            axis.plot(np.cos(phi), np.sin(phi), **sphere_settings)
        elif dim == 3:
            phi = np.linspace(0, 2 * np.pi, 200)
            theta = np.linspace(0, np.pi, 200)
            x = np.outer(np.cos(phi), np.sin(theta))
            y = np.outer(np.sin(phi), np.sin(theta))
            z = np.outer(np.ones_like(phi), np.cos(theta))
            axis.plot_surface(x, y, z, **sphere_settings)
    return fig, axis


def _plot_simplex(axis, simplex, **kwargs):
    for start, end in itertools.combinations(simplex, 2):
        values = list(np.array([start, end]).T)
        axis.plot(*values, **kwargs)



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the SimplexQueue, which tracks the state of simplices to be minimized.
"""

from abc import ABC, abstractmethod
from queue import Queue

import numpy as np
from fsc.export import export
from fsc.hdf5_io import HDF5Enabled, subscribe_hdf5


class ObjectQueue(HDF5Enabled, ABC):
    """
    General queue class. Implements caching (queueing objects only once) and
    HDF5 serialization on top of the built-in Queue.
    """

    HDF5_ATTRIBUTES = ['objects']

    def __init__(self, objects=frozenset()):
        all_objects = self.normalize(objects)
        self._queued_objects = Queue()
        self._extend_queue(all_objects)

        self._all_objects = set(all_objects)
        self.needs_saving = True

    @abstractmethod
    def normalize(self, objects):
        raise NotImplementedError

    @property
    def objects(self):
        return list(self._queued_objects.queue)

    def pop_queued(self):
        return self._queued_objects.get_nowait()

    def add_objects(self, objects):
        """
        Add new objects to the queue.
        """
        new_objects = self.normalize(objects)
        new_objects_filtered = [
            obj for obj in new_objects if obj not in self._all_objects
        ]
        if new_objects_filtered:
            self._extend_queue(new_objects_filtered)
            self._all_objects.update(new_objects_filtered)
            self.needs_saving = True

    def _extend_queue(self, objects):
        """
        Add given objects to '_queued_objects'. Note that this does _not_
        handle the other attributes, use 'add_objects' for this purpose.
        """
        for obj in objects:
            self._queued_objects.put_nowait(obj)

    @property
    def has_queued(self):
        """
        Shows if there are currently queued objects.
        """
        return not self._queued_objects.empty()

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        # try:
        objects = np.array(hdf5_handle['objects'])
        # except
        return cls(objects=objects)

    def to_hdf5(self, hdf5_handle):
        objects = np.array(self.objects)
        hdf5_handle['objects'] = objects


class RunningQueue(ObjectQueue, HDF5Enabled):  # pylint: disable=abstract-method
    """
    Queue class for objects which can have a 'running' state. Objects are
    automatically put in the 'running' state when pop-ed from the queue, and
    need to be set to 'finished' to be removed from the queue. When reloading
    the queue, all running objects are put back into the queue.
    """
    def __init__(self, objects=frozenset()):
        super().__init__(objects=objects)
        self._running_objects = set()

    @property
    def objects(self):
        # Give running objects first, so that they will be re-queued first when
        # restarting a calculation.
        return list(self._running_objects) + super().objects

    def pop_queued(self):
        """
        Get a queued object, and add it to the running objects.
        """
        obj = super().pop_queued()
        self._running_objects.add(obj)
        return obj

    def set_finished(self, obj):
        """
        Mark a given object as finished.
        """
        self._running_objects.remove(obj)
        self.needs_saving = True

    @property
    def finished(self):
        """
        Indicates whether the queue is finished.
        """
        return not self.objects

    @property
    def num_running(self):
        """
        Gives the number of currently running objects.
        """
        return len(self._running_objects)


@export
@subscribe_hdf5('nodefinder.simplex_queue')
class SimplexQueue(RunningQueue):
    """
    Queue class for the simplices which should be minimized.
    """
    @staticmethod
    def normalize(objects):
        return [
            tuple(sorted(tuple(coord) for coord in simplex))
            for simplex in objects
        ]


@export
@subscribe_hdf5('nodefinder.position_queue')
class PositionQueue(ObjectQueue):
    """
    Queue class for the positions which should be refined.
    """
    @staticmethod
    def normalize(objects):
        return [tuple(pos) for pos in objects]



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines a helper function for generating a starting mesh.
"""

import itertools

import numpy as np


def _generate_mesh_simplices(
    *, limits, mesh_size, periodic=False, skip_origin=False
):
    """
        Generate the starting simplices for given limits and mesh size.
        """
    dim = len(limits)
    if len(mesh_size) != dim:
        raise ValueError(
            "Inconsistent dimensions: {}, {}".format(dim, len(mesh_size))
        )
    vertices = list(
        itertools.product(
            *[
                np.linspace(lower, upper, m, endpoint=not periodic)
                for (lower, upper), m in zip(limits, mesh_size)
            ]
        )
    )
    if skip_origin:
        vertices = [v for v in vertices if not np.allclose(v, 0)]
    size = np.array([upper - lower for lower, upper in limits])
    simplex_distances = size / (2 * np.array(mesh_size))
    simplex_stencil = np.zeros(shape=(dim + 1, dim))
    for i, dist in enumerate(simplex_distances):
        simplex_stencil[i + 1][i] = dist
    return [v + simplex_stencil for v in vertices]



# ******NOTICE***************
# optimize.py module by Travis E. Oliphant
#
# You may copy and use this module as you see fit with no
# guarantee implied provided you keep this notice in all copies.
# *****END NOTICE************
#
# The additional license terms given in ADDITIONAL_TERMS.txt apply to this
# file.

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>

# pylint: skip-file

import asyncio
import warnings
import itertools

import numpy as np
from fsc.export import export

from ..result._minimization import MinimizationResult

# standard status messages of optimizers
_status_message = {
    'success':
    'Optimization terminated successfully.',
    'maxfev':
    'Maximum number of function evaluations has '
    'been exceeded.',
    'maxiter':
    'Maximum number of iterations has been '
    'exceeded.',
    'pr_loss':
    'Desired error not necessarily achieved due '
    'to precision loss.',
    'fprime_cutoff':
    'Cutoff for the maximum estimated derivative'
    'has been exceeded.',
}


def wrap_function(function):
    ncalls = [0]

    async def function_wrapper(*args):
        ncalls[0] += 1
        return await function(*args)

    return ncalls, function_wrapper


@export
async def root_nelder_mead(
    func,
    *,
    initial_simplex,
    xtol,
    ftol,
    maxiter=None,
    maxfev=None,
    fprime_cutoff=None,
    keep_history=True
):
    """
    Minimization of scalar function of one or more variables using the
    Nelder-Mead algorithm.

    The algorithm is modified for root-finding by an additional cutoff criterion,
    aborting the minimization when the function value divided by the longest
    edge of the simplex is larger than a given cutoff, ``fprime_cutoff``. This
    is an estimate for the maximum value of the first derivative, such that a
    root can be in the vicinity of the current simplex. The purpose of this
    criterion is to avoid spending a lot of effort finding local minima which
    are not roots.

    Arguments
    ---------
    initial_simplex : numpy.ndarray
        Coordinates of the initial simplex.
    xtol : float
        Relative error in solution `xopt` acceptable for convergence.
    ftol : float
        Relative error in ``fun(xopt)`` acceptable for convergence.
    maxiter : int
        Maximum number of iterations to perform.
    maxfev : int
        Maximum number of function evaluations to make.
    fprime_cutoff:
        Cutoff for the additional root-finding aborting criterion.

    Returns
    -------
    MinimizationResult:
        The result of the optimization.
    """
    maxfun = maxfev

    fcalls, func = wrap_function(func)
    N = len(initial_simplex[0])
    if maxiter is None:
        maxiter = N * 200
    if maxfun is None:
        maxfun = N * 200

    rho = 1
    chi = 2
    psi = 0.5
    sigma = 0.5
    one2np1 = list(range(1, N + 1))

    sim = np.array(initial_simplex)
    assert sim.shape == (N + 1, N)

    fsim = np.array(await asyncio.gather(*[func(x) for x in sim]), dtype=float)
    assert fsim.shape == (N + 1, )

    ind = np.argsort(fsim)
    fsim = np.take(fsim, ind, 0)
    # sort so sim[0,:] has the lowest function value
    sim = np.take(sim, ind, 0)

    # if keep_history is false, we still need the last simplex
    simplex_history = [np.copy(sim)]
    if keep_history:
        fun_simplex_history = [np.copy(fsim)]

    iterations = 1

    while (fcalls[0] < maxfun and iterations < maxiter):
        if (
            fprime_cutoff is not None
            and _get_fprime_estimate(sim=sim, fval=fsim[0]) > fprime_cutoff
        ):
            break
        with warnings.catch_warnings():
            # Ignore subtraction 'inf - inf' in fsim, since it will correctly
            # evaluate to False.
            warnings.simplefilter('ignore')
            if (
                np.max(np.ravel(np.abs(sim[1:] - sim[0]))) <= xtol
                and np.max(np.abs(fsim[0] - fsim[1:])) <= ftol
            ):
                break

        xbar = np.add.reduce(sim[:-1], 0) / N
        xr = (1 + rho) * xbar - rho * sim[-1]
        fxr = await func(xr)
        doshrink = 0

        if fxr < fsim[0]:
            xe = (1 + rho * chi) * xbar - rho * chi * sim[-1]
            fxe = await func(xe)

            if fxe < fxr:
                sim[-1] = xe
                fsim[-1] = fxe
            else:
                sim[-1] = xr
                fsim[-1] = fxr
        else:  # fsim[0] <= fxr
            if fxr < fsim[-2]:
                sim[-1] = xr
                fsim[-1] = fxr
            else:  # fxr >= fsim[-2]
                # Perform contraction
                if fxr < fsim[-1]:
                    xc = (1 + psi * rho) * xbar - psi * rho * sim[-1]
                    fxc = await func(xc)

                    if fxc <= fxr:
                        sim[-1] = xc
                        fsim[-1] = fxc
                    else:
                        doshrink = 1
                else:
                    # Perform an inside contraction
                    xcc = (1 - psi) * xbar + psi * sim[-1]
                    fxcc = await func(xcc)

                    if fxcc < fsim[-1]:
                        sim[-1] = xcc
                        fsim[-1] = fxcc
                    else:
                        doshrink = 1

                if doshrink:
                    for j in one2np1:
                        sim[j] = sim[0] + sigma * (sim[j] - sim[0])
                        fsim[j] = await func(sim[j])

        ind = np.argsort(fsim)
        sim = np.take(sim, ind, 0)
        fsim = np.take(fsim, ind, 0)
        iterations += 1
        if keep_history:
            simplex_history.append(np.copy(sim))
            fun_simplex_history.append(np.copy(fsim))
        else:
            simplex_history[0] = np.copy(sim)

    x = sim[0]
    fval = np.min(fsim)
    warnflag = 0

    if fcalls[0] >= maxfun:
        warnflag = 1
        msg = _status_message['maxfev']
    elif iterations >= maxiter:
        warnflag = 2
        msg = _status_message['maxiter']
    elif (
        fprime_cutoff is not None
        and _get_fprime_estimate(sim=sim, fval=fval) > fprime_cutoff
    ):
        warnflag = 3
        msg = _status_message['fprime_cutoff']
    else:
        msg = _status_message['success']

    if keep_history:
        hist_kwargs = dict(
            simplex_history=np.array(simplex_history),
            fun_simplex_history=np.array(fun_simplex_history)
        )
    else:
        hist_kwargs = dict(simplex_history=np.array(simplex_history))
    result = MinimizationResult(
        pos=x,
        value=fval,
        num_iter=iterations,
        num_fev=fcalls[0],
        status=warnflag,
        success=(warnflag == 0),
        message=msg,
        **hist_kwargs
    )
    return result


def _get_fprime_estimate(sim, fval):
    return fval / np.sqrt(
        np.max(
            np.sum(
                np.square([
                    p2 - p1 for p1, p2 in itertools.combinations(sim, r=2)
                ]),
                axis=-1
            )
        )
    )



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Submodule defining the result classes of the search step.
"""

from ._minimization import *
from ._search_result_container import *
from ._controller_state import *

__all__ = _minimization.__all__ + _search_result_container.__all__ + _controller_state.__all__  # pylint: disable=undefined-variable



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""Usage: pip install .[dev]"""

import re
import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 6):
    raise 'Must use Python version 3.6 or higher.'

with open('./README.md', 'r') as f:
    README = f.read()

with open('./nodefinder/__init__.py', 'r') as f:
    MATCH_EXPR = "__version__[^'\"]+(['\"])([^'\"]+)"
    VERSION = re.search(MATCH_EXPR, f.read()).group(2).strip()

EXTRAS_REQUIRE = dict(
    test=['pytest', 'pytest-cov', 'pytest-score'],
    doc=['sphinx', 'sphinx-rtd-theme', 'ipython>=6.2'],
    dev=[
        'pylint==2.4.4',
        'pre-commit==2.0.0',
        'prospector==1.2.0',
        'yapf==0.29',
    ],
)
EXTRAS_REQUIRE['dev'] += EXTRAS_REQUIRE['doc'] + EXTRAS_REQUIRE['test']

setup(
    name='nodefinder',
    version=VERSION,
    author='Dominik Gresch',
    author_email='greschd@gmx.ch',
    url='https://nodefinder.greschd.ch',
    description=
    'A tool for studying the nodal features of potential lanscapes.',
    install_requires=[
        'numpy', 'scipy', 'matplotlib', 'decorator', 'fsc.export',
        'fsc.hdf5-io~=1.0', 'fsc.async_tools', 'networkx>=2.0'
    ],
    python_requires=">=3.6",
    extras_require=EXTRAS_REQUIRE,
    long_description=README,
    long_description_content_type="text/markdown",
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English', 'Operating System :: Unix',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'Development Status :: 4 - Beta'
    ],
    license='Apache 2.0',
    keywords=['minimization', 'node search'],
    entry_points={'fsc.hdf5_io.load': ['nodefinder = nodefinder']},
    packages=find_packages()
)



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the top-level result class for the search step.
"""

import numpy as np
from fsc.export import export
from fsc.hdf5_io import SimpleHDF5Mapping, subscribe_hdf5

from ._cell_list import CellList


@export
@subscribe_hdf5(
    'nodefinder.search_result_container',
    extra_tags=['nodefinder.result_container']
)
class SearchResultContainer(SimpleHDF5Mapping):
    """
    Container for the results of a search run.

    Attributes
    ----------
    coordinate_system : CoordinateSystem
        Coordinate system used.
    nodes : list(MinimizationResult)
        Minimization results which fulfill the gap threshold criterion.
    gap_threshold : float
        Threshold for results to be considered a node.
    dist_cutoff : float
        Cutoff distance for searching neighbouring nodes.

    """

    HDF5_ATTRIBUTES = [
        'coordinate_system',
        'minimization_results',
        'dist_cutoff',
        'gap_threshold',
    ]
    HDF5_OPTIONAL = ['refined_results']

    def __init__(
        self,
        *,
        coordinate_system,
        minimization_results=(),
        gap_threshold,
        dist_cutoff,
        refined_results=()
    ):
        self.coordinate_system = coordinate_system
        self.gap_threshold = gap_threshold
        self.dist_cutoff = dist_cutoff

        if dist_cutoff == 0:
            num_cells = np.full_like(self.coordinate_system.size, 100)
        else:
            num_cells = np.minimum(  # pylint: disable=assignment-from-no-return,useless-suppression
                100,
                np.maximum(
                    1,
                    np.array(
                        self.coordinate_system.size / self.dist_cutoff,
                        dtype=int
                    )
                )
            )
        self.nodes = CellList(
            num_cells=num_cells, periodic=self.coordinate_system.periodic
        )
        self.rejected_results = []
        for res in minimization_results:
            self.add_result(res)
        self.refined_results = CellList(
            num_cells=num_cells, periodic=self.coordinate_system.periodic
        )
        for res in refined_results:
            self.set_refined(res)
        self.needs_saving = True

    def __repr__(self):
        return 'SearchResultContainer(coordinate_system={0.coordinate_system}, minimization_results=<{1} values>, gap_threshold={0.gap_threshold!r}, dist_cutoff={0.dist_cutoff!r})'.format(
            self, len(self.minimization_results)
        )

    def add_result(self, res):
        """
        Add a minimization result to the container.

        Arguments
        ---------
        res : MinimizationResult
            Minimization result to add.
        """
        res.pos = self.coordinate_system.normalize_position(res.pos)
        if not res.success or res.value > self.gap_threshold:  # pylint: disable=no-else-return
            self.rejected_results.append(res)
            return False
        else:
            self.nodes.add_point(self.coordinate_system.get_frac(res.pos), res)
            return True
        self.needs_saving = True

    def set_refined(self, pos):
        """
        Set a position to be refined.

        Arguments
        ---------
        pos : np.array
            The position from where refinement started.
        """
        self.refined_results.add_point(
            self.coordinate_system.get_frac(pos), pos
        )
        self.needs_saving = True

    @property
    def minimization_results(self):
        """
        list(MinimizationResult):
            All minimization results, including rejected points.
        """
        return self.nodes.values() + self.rejected_results

    def _get_neighbour_iterator(self, pos):
        candidates = self.nodes.get_neighbour_values(
            frac=self.coordinate_system.get_frac(pos)
        )
        return (c for c in candidates if np.any(c.pos != pos))

    def get_neighbour_distance_iterator(self, pos):
        """
        Returns an iterator over the distance to neighbouring nodes from a given
        position. Only neighbours within ``dist_cutoff`` are taken into account.

        Arguments
        ---------
        pos : numpy.ndarray
            Position for which to calculate the distances.
        """
        candidates = self._get_neighbour_iterator(pos)
        return (
            self.coordinate_system.distance(pos, c.pos) for c in candidates
        )

    def get_refined_neighbour_distance_iterator(self, pos):  # pylint: disable=invalid-name
        """
        Returns an iterator over the distance to neighboring nodes which have
        been used as a starting point in a refinement procedure.

        Arguments
        ---------
        pos : numpy.ndarray
            Position for which to calculate the distances.
        """
        candidates = self.refined_results.get_neighbour_values(
            frac=self.coordinate_system.get_frac(pos)
        )
        return (self.coordinate_system.distance(pos, c) for c in candidates)

    def get_all_neighbour_distances(self, pos):
        """
        Calculate the distances to neighbouring nodes from a given position.
        Only neighbours within ``dist_cutoff`` are taken into account.

        Arguments
        ---------
        pos : numpy.ndarray
            Position for which to calculate the distances.
        """
        candidates = self._get_neighbour_iterator(pos)
        positions = np.array([c.pos for c in candidates])

        if positions.size == 0:
            return []
        return self.coordinate_system.distance(pos, positions)



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the cell list class, used for quicker lookup of the neighbours of a
given position.
"""

import itertools

import numpy as np
from fsc.export import export


@export
class CellList:
    """
    Cell list container for the NodalPoint objects.
    """
    def __init__(self, num_cells, periodic):
        self.periodic = periodic
        self.num_cells = np.array(num_cells, dtype=int)
        self._total_num_cells = np.copy(self.num_cells)
        if not self.periodic:
            self._total_num_cells += 2  # add 'boundary' boxes for outside points.
        assert np.all(self.num_cells > 0)
        self._cells = np.empty(shape=self._total_num_cells, dtype=np.object)
        filler = np.frompyfunc(lambda x: list(), 1, 1)
        filler(self._cells, self._cells)
        self._values_flat = []

        self._neighbour_offset = np.array(
            list(itertools.product([-1, 0, 1], repeat=len(self.num_cells)))
        )
        self._neighbour_indices = dict()

    def _get_neighbour_indices(self, idx):
        """
        Get the indices of neighbouring cells for a given index.
        """
        try:
            return self._neighbour_indices[idx]
        except KeyError:
            res = self._calculate_neighbour_indices(idx)
            self._neighbour_indices[idx] = res
            return res

    def _calculate_neighbour_indices(self, idx):  # pylint: disable=missing-function-docstring
        indices = idx + self._neighbour_offset

        if self.periodic:
            return [tuple(i % self._total_num_cells) for i in indices]
        return [
            tuple(i) for i in indices
            if np.all(i >= 0) and np.all(i < self._total_num_cells)
        ]

    def add_point(self, frac, value):
        idx = self.get_index(frac)
        self._cells[idx].append(value)
        self._values_flat.append(value)

    def get_index(self, frac):  # pylint: disable=missing-function-docstring
        vals = np.array(frac * self.num_cells, dtype=int)
        if not self.periodic:
            vals += 1
            vals = np.maximum(0, np.minimum(vals, self._total_num_cells - 1))  # pylint: disable=assignment-from-no-return,useless-suppression
        return tuple(vals)

    def values(self):
        return self._values_flat

    def __len__(self):
        return len(self._values_flat)

    def __iter__(self):
        return iter(self._values_flat)

    def __getitem__(self, key):
        return self._values_flat[key]

    def get_neighbour_values(self, frac):
        idx = self.get_index(frac)
        for cell_idx in self._get_neighbour_indices(idx):
            yield from self._cells[cell_idx]



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the fake potential class.
"""


class FakePotential:
    """
    Defines the fake potential used to repel the minimization from the existing
    nodes.

    Arguments
    ---------
    result : SearchResultContainer
        The existing results from which the minimization should be repelled.
    width : float
        Distance from existing nodes at which the fake potential should start.
    """
    def __init__(self, result, width):
        self.result = result
        self.width = width

    def __call__(self, pos):  # pylint: disable=missing-function-docstring
        if any(
            dist < self.width
            for dist in self.result.get_all_neighbour_distances(pos)
        ):
            return float('inf')
        return 0



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the function for evaluating the dimension of a cluster of positions.
"""

import random
import warnings
import itertools
from collections import Counter

import numpy as np
import scipy.linalg as la
from scipy.special import binom
from fsc.export import export

EXACT_VALUES = {1: 0.5, 2: 8 / (9 * np.pi), 3: 27 * np.pi / 512}


@export
def calculate_dimension(
    *,
    graph,
    feature_size,
    coordinate_system,
    max_dim=3,
    min_pos_evaluations=5,
    min_neighbour_evaluations=10
):
    """
    Calculate the dimension of a cluster of positions.

    Arguments
    ---------
    positions : list(tuple(float))
        The positions whose dimension should be evaluated.
    feature_size : float
        Distance between two nodal points at which they are considered to be
        distinct. This is used as a characteristic length scale to determine
        the dimension.
    coordinate_system : CoordinateSystem
        Coordinate system used to calculate distances.
    max_dim : int
        Upper limit for the dimension.
    min_pos_evaluations : int
        Minimum number of positions which are evaluated to estimate the dimension.
    min_neighbour_evaluations : int
        Minimum number of neighbour tuples which are evaluated per position.
    """
    positions_to_evaluate = list(graph.nodes)

    if min_pos_evaluations is not None:
        random.shuffle(positions_to_evaluate)

    dim_counter = Counter()
    for i, pos in enumerate(positions_to_evaluate):
        if i >= min_pos_evaluations:
            finished, dim = _check_count(dim_counter)
            if finished:
                return dim
        dim_counter.update([
            _get_dimension(
                pos,
                coordinate_system=coordinate_system,
                graph=graph,
                feature_size=feature_size,
                max_dim=max_dim,
                min_neighbour_evaluations=min_neighbour_evaluations
            )
        ])
    finished, dim = _check_count(dim_counter)
    if not finished:
        warnings.warn('Inconclusive dimension count: {}'.format(dim_counter))
        return None
    return dim


def _check_count(counter):
    """
    Get the most common dimension, and check if at least two thirds of the
    values agree.
    """
    dim, count = counter.most_common(1)[0]
    if count >= (2 / 3) * sum(counter.values()):
        return True, dim
    else:
        return False, dim


def _get_dimension(
    pos, *, graph, coordinate_system, feature_size, max_dim,
    min_neighbour_evaluations
):
    """
    Get the dimension from a given position.
    """
    for dim in range(1, max_dim + 1):
        if not _has_dimension(
            pos=pos,
            graph=graph,
            coordinate_system=coordinate_system,
            dim=dim,
            min_neighbour_evaluations=min_neighbour_evaluations,
            feature_size=feature_size
        ):
            return dim - 1
    return max_dim


def _has_dimension(
    pos, *, graph, coordinate_system, dim, min_neighbour_evaluations,
    feature_size
):
    """
    Check if a position has at least the given dimension.
    """
    if len(pos) < dim:
        return 0
    neighbours = list(graph.neighbors(pos))
    if min_neighbour_evaluations is None:
        min_neighbour_evaluations = len(neighbours)

    try:
        limit_value = feature_size**dim * EXACT_VALUES[dim] / 2
    except KeyError:
        raise NotImplementedError(
            'Getting the {}-dimensional volume is not implemented'.format(dim)
        )

    def draw_neighbour_tuple():
        """
        Generator to draw a tuple of neighbors, without repeating.
        """
        used_tuples = set()
        num_neighbours = len(neighbours)
        num_combinations = binom(num_neighbours, dim)
        # random sampling with check for uniqueness -- efficient for very
        # large number of possible combinations
        if num_combinations > 10 * min_neighbour_evaluations:
            while len(used_tuples) < (num_combinations / 2):
                val = tuple(sorted(random.sample(neighbours, k=dim)))

                if val not in used_tuples:
                    yield val
                    used_tuples.add(val)

        # switch to directly sampling from the remaining combinations
        # efficient when there are not so many (remaining) combinations
        neighbour_tuples = set(
            tuple(sorted(n))
            for n in itertools.combinations(neighbours, r=dim)
        ) - used_tuples
        num_draws = 10
        while neighbour_tuples:
            vals = random.sample(
                neighbour_tuples, min(num_draws, len(neighbour_tuples))
            )
            yield from vals
            neighbour_tuples -= set(vals)

    results = []
    for i, neighbour_tuple in enumerate(draw_neighbour_tuple()):
        if i >= min_neighbour_evaluations and i > 1:
            avg = np.average(results)
            error_mean = np.sqrt(np.var(results) / (i - 1))
            if abs(limit_value - avg) / 2 > error_mean:
                return avg > limit_value
        results.append(
            _get_volume(
                pos=pos,
                neighbour_tuple=neighbour_tuple,
                coordinate_system=coordinate_system
            )
        )
    return np.average(results) > limit_value


def _get_volume(pos, neighbour_tuple, coordinate_system):
    """
    Get the volume spanned by a position and the given neighbours.
    """
    connecting_vectors = [
        coordinate_system.connecting_vector(
            np.array(neighbour), np.array(pos)
        ) for neighbour in neighbour_tuple
    ]
    mat = np.array(connecting_vectors)
    svd = la.svd(mat, compute_uv=False)
    if len(svd) < len(neighbour_tuple):
        return 0
    return abs(np.product(svd))



# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Contains functions for creating the refinement stencil used by the search
procedure.
"""

import numpy as np
import scipy.linalg as la

from fsc.export import export

from ._mesh_helper import _generate_mesh_simplices


@export
def get_mesh_stencil(*, mesh_size, dist_multiplier=2.5):
    """
    Create a mesh refinement stencil.

    Arguments
    ---------
    mesh_size : list(int)
        The size of the mesh in each dimension.
    dist_multiplier : float
        Defines the size of the refinement mesh. A multiplier of one means that
        the refinement box extends to the ``dist_cutoff`` on each side.
    """
    limits = [(-dist_multiplier, dist_multiplier)] * len(mesh_size)
    return np.array(
        _generate_mesh_simplices(
            mesh_size=mesh_size, limits=limits, skip_origin=True
        )
    )


@export
def get_auto_stencil(*, dim):
    """
    Get the default stencil for a given dimension.

    Arguments
    ---------
    dim : int
        The problem dimension.
    """
    if dim == 2:
        return get_circle_stencil(num_points=5)
    elif dim == 3:
        return get_sphere_stencil(num_points=30)
    return get_mesh_stencil(mesh_size=[3] * dim)


def get_circle_stencil(*, num_points):
    """
    Produce a stencil with simplices along a circle. Only suitable for
    two-dimensional problems.

    Arguments
    ---------
    num_points : int
        The number of points on the circle / number of simplices.
    """
    phi = np.linspace(0, 2 * np.pi, num_points, endpoint=False)
    simplex = (
        np.array([[0, 0], [1 / 2, np.sqrt(3) / 2],
                  [-1 / 2, np.sqrt(3) / 2]]) * 5 / num_points
    ) + [0, 1.5]
    res = np.zeros((num_points, 3, 2))
    for i, phi_val in enumerate(phi):
        res[i, :, :] = (
            np.array([[np.cos(phi_val), -np.sin(phi_val)],
                      [np.sin(phi_val), np.cos(phi_val)]]) @ simplex.T
        ).T
    return res


def get_sphere_stencil(*, num_points):
    """
    Produce a stencil with simplices on the surface of a sphere. Only suitable
    for three-dimensional problems.

    Arguments
    ---------
    num_points : int
        The number of simplices which are placed on the sphere.
    """
    points = 1.1 * np.array(_fibonacci_sphere_points(num_points))
    simplex_edge_length = 3 / np.sqrt(num_points)
    simplex = np.zeros((4, 3))
    simplex[1:, :] = (0.25 + 0.75 * np.eye(3)) * simplex_edge_length
    q_mat_1, r_mat_1 = la.qr([[1, 0, 0], [1, 0, 0], [1, 0, 0]])
    q_mat_1 *= np.sign(r_mat_1[0, 0])
    simplex = (q_mat_1 @ simplex.T).T

    res = []
    for pos in points:
        mat = np.zeros((3, 3))
        mat[:, 0] = pos
        q_mat_2, r_mat_2 = la.qr(mat)
        q_mat_2 *= np.sign(r_mat_2[0, 0])
        res.append((q_mat_2 @ simplex.T).T + pos)

    return np.array(res)


def _fibonacci_sphere_points(num_points):
    """
    Helper function that places points on a sphere using the Fibonacci spiral.
    """
    res = []
    offset = 2 / num_points
    increment = np.pi * (3 - np.sqrt(5))

    for i in range(num_points):
        z = (i + 0.5) * offset - 1
        rho = np.sqrt(1 - z**2)
        phi = ((i + 1) % num_points) * increment

        x = np.cos(phi) * rho
        y = np.sin(phi) * rho

        res.append([x, y, z])
    return res
